from django.shortcuts import get_object_or_404
from rest_framework.generics import \
    ListAPIView, ListCreateAPIView, RetrieveUpdateDestroyAPIView, RetrieveAPIView
from rest_framework.views import APIView
from rest_framework.parsers import MultiPartParser, FileUploadParser
from rest_framework.permissions import IsAuthenticated, IsAdminUser
from rest_framework.response import Response
from rest_framework import status
from rest_framework.exceptions import ValidationError
from arkindex.documents.models import Corpus
from arkindex.dataimport.models import \
    DataImport, DataFile, DataImportState, DataImportMode, DataImportFailure, Repository
from arkindex.dataimport.serializers import \
    DataImportLightSerializer, DataImportSerializer, DataImportFailureSerializer, DataFileSerializer, \
    RepositoryLightSerializer, RepositorySerializer, ExternalRepositorySerializer
from arkindex.users.models import OAuthCredentials
import hashlib
import magic


class DataImportsList(ListCreateAPIView):
    """
    List data imports
    """
    permission_classes = (IsAuthenticated, )
    serializer_class = DataImportLightSerializer

    def get_queryset(self):
        return DataImport.objects.filter(creator=self.request.user)

    def perform_create(self, serializer):
        # Temporary limitation
        if serializer.validated_data['mode'] not in (DataImportMode.Images, ):
            raise ValidationError('Unsupported mode for now, sorry.')

        return super().perform_create(serializer)


class DataImportDetails(RetrieveUpdateDestroyAPIView):
    """
    Retrieve and edit a data import
    """

    permission_classes = (IsAuthenticated, )
    serializer_class = DataImportSerializer

    def get_queryset(self):
        return DataImport.objects.filter(creator=self.request.user)

    def perform_update(self, serializer):
        dataimport = serializer.instance
        if dataimport.state not in (DataImportState.Created, DataImportState.Configured):
            raise ValidationError("Cannot update a workflow after it has been started")

        # Allow updating the state from Configured to Running only
        if dataimport.state != serializer.validated_data['state']:
            if dataimport.state != DataImportState.Configured:
                raise ValidationError({'state': ['Cannot change state on non-configured or started workflows']})
            if serializer.validated_data['state'] != DataImportState.Running:
                raise ValidationError(
                    {'state': ['The state of a configured workflow can only be set to configured or running']})
            dataimport.start()
            return

        super().perform_update(serializer)

        # Automatically change from Created to Configured if a payload has been set
        if dataimport.state == DataImportState.Created and dataimport.payload is not None:
            dataimport.state = DataImportState.Configured
            dataimport.save()

    def perform_destroy(self, instance):
        if instance.state == DataImportState.Running:
            raise ValidationError("Cannot delete a workflow while it is running")
        super().perform_destroy(instance)


class DataImportFailures(ListAPIView):
    """
    List a single import's errors
    """

    permission_classes = (IsAuthenticated, )
    serializer_class = DataImportFailureSerializer

    def get_queryset(self):
        return DataImportFailure.objects.filter(dataimport_id=self.kwargs['pk']) \
                                        .prefetch_related('dataimport__revision__repo', 'element') \
                                        .order_by('path', 'line')


class DataFileList(ListAPIView):
    """
    List uploaded files
    """
    permission_classes = (IsAuthenticated, )
    serializer_class = DataFileSerializer

    def get_queryset(self):
        return DataFile.objects.filter(corpus_id=self.kwargs['pk'])


class DataFileRetrieve(RetrieveUpdateDestroyAPIView):
    """
    Get one file
    """
    permission_classes = (IsAuthenticated, )
    serializer_class = DataFileSerializer
    # TODO: Restrict to user's corpuses
    queryset = DataFile.objects.all()


class DataFileUpload(APIView):
    """
    Upload a new file to a corpus
    """
    permission_classes = (IsAuthenticated, )
    parser_classes = (MultiPartParser, FileUploadParser)

    def post(self, request, pk=None, format=None):
        corpus_qs = Corpus.objects.filter(id=pk)
        if not corpus_qs.exists():
            raise ValidationError({'corpus': ['Corpus not found']})
        corpus = corpus_qs.get()

        file_obj = request.FILES['file']

        md5hash = hashlib.md5()
        for chunk in file_obj.chunks():
            md5hash.update(chunk)
        file_hash = md5hash.hexdigest()

        # Reopen file to reread from beginning
        file_obj.open()
        file_type = magic.from_buffer(file_obj.read(1024), mime=True)

        if DataFile.objects.filter(hash=file_hash, corpus=corpus).exists():
            raise ValidationError('File already exists')

        df = DataFile.objects.create(
            hash=file_hash,
            corpus=corpus,
            name=file_obj.name,
            size=file_obj.size,
            content_type=file_type,
        )

        file_obj.open()
        with open(df.staging_path, 'wb') as destfile:
            for chunk in file_obj.chunks():
                destfile.write(chunk)

        return Response(
            data=DataFileSerializer(df).data,
            status=status.HTTP_201_CREATED)


class GitRepositoryImportHook(APIView):
    """
    Handle Git push events
    """

    def post(self, request, pk=None, **kwargs):
        repo = get_object_or_404(Repository, id=pk)
        repo.provider_class(credentials=repo.credentials).handle_webhook(repo, request)
        return Response(status=status.HTTP_204_NO_CONTENT)


class RepositoryList(ListAPIView):
    permission_classes = (IsAuthenticated, )
    serializer_class = RepositoryLightSerializer

    def get_queryset(self):
        return Repository.objects.filter(credentials__user=self.request.user)


class AvailableRepositoriesList(ListCreateAPIView):
    permission_classes = (IsAuthenticated, )
    pagination_class = None
    serializer_class = ExternalRepositorySerializer

    def get_queryset(self):
        cred = get_object_or_404(OAuthCredentials, user=self.request.user, id=self.kwargs['pk'])
        return cred.git_provider_class(credentials=cred).list_repos(
            query=self.request.query_params.get('search'),
        )

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        cred = get_object_or_404(OAuthCredentials, user=self.request.user, id=self.kwargs['pk'])

        provider = cred.git_provider_class(credentials=cred)
        repo = provider.create_repo(**serializer.validated_data, request=self.request)
        rev, _ = provider.get_or_create_latest_revision(repo)
        dataimport = rev.start_import()

        return Response(data={'import_id': str(dataimport.id)}, status=status.HTTP_201_CREATED)


class RepositoryRetrieve(RetrieveUpdateDestroyAPIView):
    permission_classes = (IsAuthenticated, )
    serializer_class = RepositorySerializer

    def get_queryset(self):
        return Repository.objects.filter(credentials__user=self.request.user)


class RepositoryStartImport(RetrieveAPIView):
    permission_classes = (IsAdminUser, )

    def get_queryset(self):
        return Repository.objects.filter(credentials__user=self.request.user)

    def get(self, request, *args, **kwargs):
        repo = self.get_object()

        rev, _ = repo.credentials.git_provider_class(
            credentials=repo.credentials,
        ).get_or_create_latest_revision(repo)

        if rev.dataimports.filter(state=DataImportState.Running).exists():
            raise ValidationError("An import is already running for the latest revision")

        return Response(data={'import_id': str(rev.start_import().id)})
