from rest_framework.generics import \
    ListAPIView, RetrieveAPIView, CreateAPIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from rest_framework import status
from django.conf import settings
from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from arkindex.documents.serializers.elements import \
    ElementSerializer, CorpusSerializer, PageLightSerializer, ActSerializer, SurfaceSerializer
from arkindex.documents.serializers.transcriptions import \
    TranscriptionsSerializer, TranscriptionCreateSerializer
from arkindex.documents.serializers.search import \
    TranscriptionSearchResultSerializer, ActSearchResultSerializer
from arkindex.documents.serializers.iiif import \
    VolumeManifestSerializer, ActManifestSerializer, \
    PageAnnotationListSerializer, PageActAnnotationListSerializer, \
    SurfaceAnnotationListSerializer, TranscriptionSearchAnnotationListSerializer
from arkindex.documents.models import \
    Element, ElementType, Page, Act, Transcription, Corpus, TranscriptionType
from arkindex.documents.search import \
    search_transcriptions_post, search_acts_post, search_transcriptions_filter_post
from arkindex.documents.indexer import Indexer
from arkindex.images.models import Zone
from arkindex.images.importer import bulk_transcriptions
from arkindex.project.elastic import ESQuerySet
from arkindex.project.polygon import Polygon
from arkindex.project.tools import elasticsearch_escape


class ElementsList(ListAPIView):
    """
    List documents using pagination
    """
    serializer_class = ElementSerializer

    def get_queryset(self):
        return Element.objects.filter(
            type=self.request.query_params.get('type', ElementType.Volume)
        ).order_by('type', 'name')


class RelatedElementsList(ListAPIView):
    """
    List all documents linked to a specified document
    """
    serializer_class = ElementSerializer

    def get_queryset(self):
        filtering = {}
        type_param = self.request.query_params.get('type')
        if type_param is not None:
            filtering['type'] = type_param
        related = Element.objects.get_related(self.kwargs['pk'], **filtering)
        return related


class ElementPages(ListAPIView):
    """
    List pages inside an element
    """
    serializer_class = PageLightSerializer

    def get_queryset(self):
        return Page.objects.get_descending(self.kwargs['pk']).prefetch_related('zone__image__server')


class ElementSurfaces(ListAPIView):
    """
    List surfaces inside a volume
    """
    serializer_class = ElementSerializer

    def get_queryset(self):
        return Element.objects.get_descending(self.kwargs['pk']) \
                              .filter(type=ElementType.Surface)


class CorpusList(ListAPIView):
    """
    List all available corpora
    """
    serializer_class = CorpusSerializer
    pagination_class = None
    queryset = Corpus.objects.all().order_by('name')


class SurfaceDetails(RetrieveAPIView):
    """
    Get details for a specific surface
    """
    serializer_class = SurfaceSerializer
    queryset = Element.objects.filter(type=ElementType.Surface) \
                              .select_related('zone__image__server')

    def get_object(self):
        surf = super().get_object()
        surf.page = Page.objects.filter(zone__image=surf.zone.image) \
                                .select_related('zone__image__server').first()
        return surf


# TODO: Edition disabled
class ActEdit(RetrieveAPIView):
    """
    Edit acts and associated ElementPaths
    """
    serializer_class = ActSerializer
    queryset = Act.objects.all()

    def get_object(self):
        act = super().get_object()
        act.parent_elements = Element.objects.get_ascending_paths(act.id)
        act.child_elements = Element.objects.get_descending(act.id)
        return act


class TranscriptionBulk(CreateAPIView):
    '''
    Create transcriptions in bulk, all linked to the same image
    and parent element
    '''
    serializer_class = TranscriptionsSerializer
    permission_classes = (IsAuthenticated, )

    def perform_create(self, serializer):

        parent = serializer.validated_data['parent']
        image = serializer.validated_data['image']
        transcriptions = bulk_transcriptions(
            image=image,
            parent=parent,
            items=[
                {
                    'polygon': tr['polygon'],
                    'text': tr['text'],
                    'score': tr['score'],
                    'type': tr['type'],
                }
                for tr in serializer.validated_data['transcriptions']
            ],
        )

        # Reindex all created
        if transcriptions:
            Indexer().run_index(
                settings.ES_INDEX_TRANSCRIPTIONS,
                Transcription.INDEX_TYPE,
                Transcription.objects.filter(
                    id__in=[t.id for t in transcriptions]
                ),
            )


class TranscriptionCreate(CreateAPIView):
    serializer_class = TranscriptionCreateSerializer
    permission_classes = (IsAuthenticated, )

    def perform_create(self, serializer):
        transcription_type = serializer.validated_data['type']
        if transcription_type not in TranscriptionType:
            raise ValidationError("This endpoint can only import transcriptions.")
        element = serializer.validated_data['element']

        # Create a zone on the page's image
        polygon = Polygon(serializer.validated_data['polygon'])
        try:
            ts_zone, _ = Zone.objects.get_or_create(
                image=element.zone.image,
                polygon=polygon,
            )
        except Zone.MultipleObjectsReturned:
            # This should not happend, but we cannot set a unique
            # on polygon PG fields
            ts_zone = Zone.objects.filter(
                image=element.zone.image,
                polygon=polygon,
            ).first()

        ts, created = Transcription.objects.get_or_create(
            element=element,
            type=transcription_type,
            zone=ts_zone,
            text=serializer.validated_data['text'],
            defaults={
                'score': serializer.validated_data['score']
            }
        )
        if not created or ts.score != serializer.validated_data['score']:
            ts.score = serializer.validated_data['score']
            ts.save()

        # Index in ES
        Indexer().run_index(
            settings.ES_INDEX_TRANSCRIPTIONS,
            Transcription.INDEX_TYPE,
            Transcription.objects.filter(id=ts.id)
        )
        return ts

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        obj = self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response({'id': obj.id}, status=status.HTTP_201_CREATED, headers=headers)


class SearchAPIMixin(object):
    def get(self, request, *args, **kwargs):
        q = request.query_params.get('q')
        if not q or q.isspace():
            raise ValidationError('A search query is required')
        return super().get(request, *args, **kwargs)


class TranscriptionSearch(SearchAPIMixin, ListAPIView):
    """
    Search and list transcriptions, using pagination
    """
    serializer_class = TranscriptionSearchResultSerializer

    def get_queryset(self):
        query = self.request.query_params.get('q')
        if not query:
            return
        corpus = self.request.query_params.get('corpus')
        type = self.request.query_params.get('type')
        return ESQuerySet(
            query=ESQuerySet.make_query(
                'elastic/search_transcriptions.json',
                ctx={'query': elasticsearch_escape(query), 'corpus': corpus, 'type': type},
            ),
            sort={"score": {"order": "desc", "mode": "max"}},
            es_index=settings.ES_INDEX_TRANSCRIPTIONS,
            es_type=Transcription.INDEX_TYPE,
            post_process=search_transcriptions_post,
        )


class ActSearch(SearchAPIMixin, ListAPIView):
    """
    Search for acts containing a specific word
    """
    serializer_class = ActSearchResultSerializer

    def get_queryset(self):
        query = self.request.query_params.get('q')
        if not query:
            return
        corpus = self.request.query_params.get('corpus')
        type = self.request.query_params.get('type')
        return ESQuerySet(
            _source=False,
            query=ESQuerySet.make_query(
                'elastic/search_acts.json',
                ctx={'query': elasticsearch_escape(query), 'corpus': corpus, 'type': type},
            ),
            es_index=settings.ES_INDEX_ACTS,
            es_type=Act.INDEX_TYPE,
            post_process=search_acts_post,
        )


class VolumeManifest(RetrieveAPIView):
    """
    Get a IIIF manifest for a specific volume
    """

    serializer_class = VolumeManifestSerializer
    queryset = Element.objects.filter(type=ElementType.Volume)

    @method_decorator(cache_page(3600))
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class ActManifest(RetrieveAPIView):
    """
    Get a IIIF manfiest for a specific act
    """

    serializer_class = ActManifestSerializer
    queryset = Act.objects.all()

    @method_decorator(cache_page(3600))
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class PageAnnotationList(RetrieveAPIView):
    """
    Get a IIIF annotation list for transcriptions of a specific page
    """

    serializer_class = PageAnnotationListSerializer
    queryset = Page.objects.all()

    @method_decorator(cache_page(3600))
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class PageActAnnotationList(RetrieveAPIView):
    """
    Get a IIIF annotation list for acts of a specific page
    """

    serializer_class = PageActAnnotationListSerializer
    queryset = Page.objects.all()

    @method_decorator(cache_page(3600))
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class SurfaceAnnotationList(RetrieveAPIView):
    """
    Get a IIIF annotation list for a single surface
    """
    serializer_class = SurfaceAnnotationListSerializer
    queryset = Element.objects.filter(type=ElementType.Surface)

    @method_decorator(cache_page(3600))
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class TranscriptionSearchAnnotationList(RetrieveAPIView):
    """
    Search for transcriptions inside an element and get a IIIF annotation list
    """

    serializer_class = TranscriptionSearchAnnotationListSerializer

    def get_object(self):
        return ESQuerySet(
            query=ESQuerySet.make_query(
                'elastic/search_transcriptions.json',
                ctx={'query': self.request.query_params.get('q'), 'type': 'word'},
            ),
            es_index=settings.ES_INDEX_TRANSCRIPTIONS,
            es_type=Transcription.INDEX_TYPE,
            post_process=search_transcriptions_filter_post,
            post_process_args=[self.kwargs['pk']]
        )
