from django.conf import settings
from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk as es_bulk
from elasticsearch.exceptions import NotFoundError
from arkindex.documents.models import Act
import logging
import time
import datetime

logger = logging.getLogger(__name__)


class Indexer(object):

    def __init__(self, hosts=settings.ELASTIC_SEARCH_HOSTS):
        self.elastic = Elasticsearch(hosts)

    def setup(self):
        """
        Create indexes in ElasticSearch
        """
        self.elastic.indices.create(
            index=settings.ES_INDEX_ACTS,
            body={
                "mappings": {
                    Act.INDEX_TYPE: {
                        "properties": {
                            "transcriptions": {
                                "type": "nested",
                                "properties": {
                                    "id": {"type": "text"},
                                    "type": {"type": "text"},
                                    "score": {"type": "float"},
                                    "text": {"type": "text"},
                                    "corpus": {"type": "text"}
                                }
                            }
                        }
                    }
                }
            }
        )

    def drop_index(self, index_name):
        """
        Try to drop an existing index
        """
        try:
            self.elastic.indices.delete(index=index_name)
            logger.info("Dropped index {}".format(index_name))
        except NotFoundError:
            logger.info("Could not drop index {} (does not exist)".format(index_name))

    def run_index(self, index_name, index_type, queryset, bulk_size=400):
        """
        Reindex elements from a QuerySet chunk by chunk and output status
        """
        count = queryset.count()
        logger.info("{} items to reindex".format(count))

        start_time, last, indexed = time.time(), 0, 0
        for i in range(0, count, bulk_size):
            percent = int(100.0 * i / count)
            if last != percent or bulk_size > 100:
                secs = time.time() - start_time
                per_second = int(i / secs)
                logger.info(
                    '{0: >3}% => {1: >5} / {2} (indexed: {3}, execution time: {4}, {5} items/s)'.format(
                        percent, i, count, indexed, datetime.timedelta(seconds=secs), per_second))
                last = percent

            indexed += self.index(index_name, index_type, queryset[i:i+bulk_size])

    def index(self, index_name, index_type, queryset):
        """
        Update the acts payload into ElasticSearch
        """

        # Build raw ElasticSearch insert
        actions = [{
            '_index': index_name,
            '_type': index_type,
            '_source': item.build_search_index(),
            '_id': item.id.hex,
        } for item in queryset.iterator()]

        # Run actions in bulk
        try:
            nb_insert, _ = es_bulk(self.elastic, actions, chunk_size=100, initial_backoff=2, max_retries=2)
        except Exception as e:
            logger.warning('Failed to bulk insert in ES : {}'.format(e))
            return 0

        return nb_insert
