#!/usr/bin/env python3
from django.core.management.base import BaseCommand
from django.conf import settings
from arkindex.documents.models import Element, ElementType
from arkindex.documents.indexer import Indexer
from arkindex.documents.tasks import reindex_acts, reindex_transcriptions
from celery import chain
import logging

logging.basicConfig(
    level=logging.INFO,
    format='[%(levelname)s] %(message)s',
)
logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'Selectively reindex transcriptions and acts into ElasticSearch'

    def add_arguments(self, parser):
        parser.add_argument(
            '--acts',
            help='Reindex acts',
            action='store_true'
        )
        parser.add_argument(
            '-ts', '--transcriptions',
            help='Reindex transcriptions',
            action='store_true'
        )
        parser.add_argument(
            '--volume',
            help='Restrict reindexing to a specific volume'
        )
        parser.add_argument(
            '--drop',
            help="Drop the existing indexes before reindexing",
            action='store_true'
        )

    def handle(self, *args, **options):
        # Handle verbosity level
        verbosity = int(options['verbosity'])
        root_logger = logging.getLogger('')
        if verbosity > 1:
            root_logger.setLevel(logging.DEBUG)

        indexer = Indexer()
        volume = None
        tasks = []

        if options['volume'] is not None:
            volume = Element.objects.get(type=ElementType.Volume, name__icontains=options['volume'])

        if options['transcriptions'] or not options['acts']:
            if options['drop']:
                indexer.drop_index(settings.ES_INDEX_TRANSCRIPTIONS)
            if volume:
                reindex_transcriptions(volume_id=volume.id)
                return
                tasks.append(reindex_transcriptions.si(volume_id=volume.id))
            else:
                tasks.append(reindex_transcriptions.si())

        if options['acts'] or not options['transcriptions']:
            if options['drop']:
                indexer.drop_index(settings.ES_INDEX_ACTS)
                indexer.setup()
            if volume:
                tasks.append(reindex_acts.si(volume_id=volume.id))
            else:
                tasks.append(reindex_acts.si())

        task = chain(*tasks).delay(async=False)
        print("Task started with ID {}".format(task.id))
