
from django.conf import settings
from rest_framework import serializers
from arkindex.documents.models import Element, ElementType, Page, Act
from arkindex.images.models import Image, Zone
from arkindex.project.tools import sslify_url, build_absolute_url
import urllib.parse


class ImageResourceManifestSerializer(serializers.BaseSerializer):
    """
    Serialize an image into a IIIF resource.
    """

    def to_representation(self, image):
        assert isinstance(image, Image)
        return {
            "@id": sslify_url(image.get_thumbnail_url(max_width=None, max_height=None)),
            "@type": "dctypes:Image",
            "height": image.height,
            "width": image.width,
            "service": {
                "@context": settings.IIIF_IMAGE_CONTEXT,
                "@id": sslify_url(image.url),
                "profile": settings.IIIF_IMAGE_SERVICE_PROFILE
            }
        }


class ImageThumbnailManifestSerializer(serializers.BaseSerializer):
    """
    Serialize an image into a IIIF manifest thumbnail
    """

    def to_representation(self, image):
        assert isinstance(image, Image)
        return {
            "@id": sslify_url(image.get_thumbnail_url()),
            "service": {
                "@context": settings.IIIF_IMAGE_CONTEXT,
                "@id": sslify_url(image.url),
                "profile": settings.IIIF_IMAGE_SERVICE_PROFILE
            }
        }


class ElementCanvasManifestSerializer(serializers.BaseSerializer):
    """
    Serialize an element's zone into a IIIF canvas
    """

    def get_other_content(self, element):
        return []

    def to_representation(self, element):
        assert isinstance(element, Element)
        zone = element.zone
        assert isinstance(zone, Zone)
        assert 'request' in self.context, "A request is required to generate absolute URLs"
        return {
            "@id": build_absolute_url(element, self.context['request'], 'api:canvas-manifest'),
            "@type": "sc:Canvas",
            "label": element.name,
            "height": zone.polygon.height,
            "width": zone.polygon.width,
            "images": [
                {
                    "@type": "oa:Annotation",
                    "resource": ImageResourceManifestSerializer(zone.image, context=self.context).data,
                    "on": build_absolute_url(element, self.context['request'], 'api:canvas-manifest'),
                    "motivation": "sc:painting"
                }
            ],
            "otherContent": self.get_other_content(element),
        }


class PageCanvasManifestSerializer(ElementCanvasManifestSerializer):
    """
    Serialize a page's zone into a IIIF canvas
    """

    def to_representation(self, page):
        assert isinstance(page, Page)
        serialized_element = super().to_representation(page)
        serialized_element['label'] = page.display_name
        return serialized_element

    def get_other_content(self, page):
        annotation_list_endpoint, annotation_list_name = \
            ("api:page-transcription-manifest", "Transcriptions") if settings.IIIF_TRANSCRIPTION_LIST \
            else ("api:page-act-manifest", "Actes")
        return [
            {
                "@id": build_absolute_url(page, self.context['request'], annotation_list_endpoint),
                "@type": "sc:AnnotationList",
                "label": annotation_list_name
            }
        ]


class ActPageCanvasManifestSerializer(PageCanvasManifestSerializer):
    """
    Serialize a page into a IIIF canvas with annotation lists for surfaces
    """

    def get_other_content(self, page):
        assert hasattr(page, 'act')
        query = self.context['request'].query_params.get('q')
        suffix = '?' + urllib.parse.urlencode({'q': query}) if query else ''
        return [
            {
                "@id": build_absolute_url(surface, self.context['request'], 'api:surface-manifest') + suffix,
                "@type": "sc:AnnotationList",
                "label": surface.name
            }
            for surface in Element.objects.get_descending(
                page.act.id,
                type=ElementType.Surface,
                zone__image_id=page.zone.image_id
            )
        ]


class ManifestSerializer(serializers.BaseSerializer):
    """
    Serialize an element into a IIIF manifest
    """

    canvas_serializer = ElementCanvasManifestSerializer
    id_url_name = 'api:volume-manifest'

    def to_representation(self, element):
        assert isinstance(element, Element)
        assert 'request' in self.context, "A request is required to generate absolute URLs"

        canvases = self.canvas_serializer(
            self.get_canvases(element),
            context=self.context,
            many=True
        ).data
        return {
            "@context": settings.IIIF_PRESENTATION_CONTEXT,
            "@id": build_absolute_url(element, self.context['request'], self.id_url_name),
            "@type": "sc:Manifest",
            "thumbnail": ImageThumbnailManifestSerializer(element.get_thumbnail()).data,
            "related": [],
            "structures": self.get_structures(element, canvases),
            "description": "",
            "sequences": [
                {
                    "canvases": canvases,
                    "label": "",
                    "@id": build_absolute_url(element, self.context['request'], 'api:sequence-manifest'),
                    "@type": "sc:Sequence"
                }
            ],
            "viewingHint": "individuals",
            "label": element.name,
            "viewingDirection": "left-to-right"
        }

    def get_canvases(self, element):
        return Element.objects.get_descending(element.id).prefetch_related('zone__image')

    def get_structures(self, element, canvases):
        return [{
            "viewingHint": "top",
            "label": element.name,
            "@id": build_absolute_url(element, self.context['request'], self.id_url_name),
            "ranges": [c['@id'] for c in canvases],
            "@type": "sc:Range"
        }] + [
            {
                "canvases": [c['@id']],
                "label": c['label'],
                "@id": c['@id'],
                "within": c['@id'],
                "@type": "sc:Range"
            }
            for c in canvases]


class VolumeManifestSerializer(ManifestSerializer):
    """
    Serialize a volume into a IIIF manifest
    """

    canvas_serializer = PageCanvasManifestSerializer
    id_url_name = 'api:volume-manifest'

    def get_canvases(self, volume):
        assert isinstance(volume, Element) and volume.type == ElementType.Volume
        return Page.objects.get_descending(volume.id).prefetch_related('zone__image__server')

    def to_representation(self, volume):
        serialized = super().to_representation(volume)
        if 'service' not in serialized:
            serialized['service'] = []
        serialized['service'].append({
            "@context": settings.IIIF_SEARCH_CONTEXT,
            "@id": build_absolute_url(volume, self.context['request'], 'api:ts-search-manifest'),
            "profile": settings.IIIF_SEARCH_SERVICE_PROFILE,
            "label": "Search transcriptions"
        })
        return serialized


class ActManifestSerializer(ManifestSerializer):
    """
    Serialize an act into a IIIF manifest
    """

    canvas_serializer = ActPageCanvasManifestSerializer
    id_url_name = 'api:act-manifest'

    def get_canvases(self, act):
        assert isinstance(act, Act)
        image_ids = list(Element.objects
                                .get_descending(act.id, type=ElementType.Surface)
                                .values_list('zone__image_id', flat=True))
        pages = Page.objects \
                    .filter(zone__image_id__in=image_ids) \
                    .select_related('zone__image__server')
        # This query gives unordered pages so we reorder them manually
        ordered_pages = sorted(pages, key=lambda p: image_ids.index(p.zone.image_id))
        # Add act info for canvas serializer
        for p in ordered_pages:
            p.act = act
        return ordered_pages
