from django.urls import reverse
from arkindex.project.tests import FixtureAPITestCase
from rest_framework import status
from arkindex.documents.models import Element, Act, ElementType, MetaData, MetaType
from arkindex.images.models import Image, Zone


class TestAct(FixtureAPITestCase):

    def setUp(self):
        # Create an act with some surfaces
        self.img = Image.objects.get(path='img6', server=self.imgsrv)
        self.act = Act.objects.create(
            corpus=self.corpus,
            name='sister act',
            number='1',
        )

        self.surfaces = []
        for i in range(3):
            x, y = i*100, i*100
            z = Zone.objects.create(polygon=[(x, y), (x+50, y), (x+50, y+50), (x, y+50), (x, y)], image=self.img)
            s = Element.objects.create(
                name='surface {}'.format(i),
                corpus=self.corpus,
                type=ElementType.Surface,
                zone=z,
            )
            s.add_parent(self.act)
            self.surfaces.append(s)

    def test_act_retrieve(self):
        '''
        Test simple GET query
        '''
        response = self.client.get(reverse('api:act-edit', kwargs={'pk': self.act.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            response.json(),
            {
                'name': 'sister act',
                'parents': [],
                'number': '1',
                'id': str(self.act.id),
                'children': [
                    {
                        'name': 'surface 0',
                        'id': str(self.surfaces[0].id),
                        'type': 'surface'
                    },
                    {
                        'name': 'surface 1',
                        'id': str(self.surfaces[1].id),
                        'type': 'surface'
                    },
                    {
                        'name': 'surface 2',
                        'id': str(self.surfaces[2].id),
                        'type': 'surface',
                    }
                ],
                'metadatas': []}
        )

    def test_act_metadatas(self):
        self.assertFalse(MetaData.objects.exists())

        # No metadatas in api
        response = self.client.get(reverse('api:act-edit', kwargs={'pk': self.act.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['metadatas'], [])

        # Create some metadatas
        metas = [
            self.act.metadatas.create(type=MetaType.Text, name='test', value='aha'),
            self.act.metadatas.create(type=MetaType.Date, name='origin', value='2010/01'),
            self.act.metadatas.create(type=MetaType.Location, name='place', value='somewhere'),
        ]

        # Retrieve them through api, sorted by name
        response = self.client.get(reverse('api:act-edit', kwargs={'pk': self.act.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            response.json()['metadatas'],
            [{'id': str(metas[1].id),
              'name': 'origin',
              'type': 'date',
              'value': '2010/01',
              'revision': None},
             {'id': str(metas[2].id),
              'name': 'place',
              'type': 'location',
              'value': 'somewhere',
              'revision': None},
             {'id': str(metas[0].id),
              'name': 'test',
              'type': 'text',
              'value': 'aha',
              'revision': None}]
        )
