from arkindex.project.tests import FixtureTestCase
from arkindex.documents.models import Transcription, Act, Element
from arkindex.documents.search import search_transcriptions_post, search_acts_post, search_transcriptions_filter_post


class TestSearchPostProcess(FixtureTestCase):
    """Test ElasticSearch post-processing functions"""

    def build_es_response(self, hits):
        return {
            "hits": {
                "total": len(hits),
                "max_score": None,
                "hits": hits
            },
            "_shards": {
                "total": 5,
                "failed": 0,
                "skipped": 0,
                "successful": 5
            },
            "took": 42,
            "timed_out": False
        }

    def make_transcription_hit(self, ts):
        return {
            "_id": str(ts.id.hex),
            "_score": None,
            "sort": [
                ts.score
            ],
            "_index": "transcriptions",
            "_type": Transcription.INDEX_TYPE,
            "_source": ts.build_search_index()
        }

    def make_nested_transcription_hit(self, ts):
        return {
            "_source": {
                "text": ts.text,
                "id": str(ts.id),
                "score": ts.score,
            },
            "_score": ts.score,
            "_nested": {
                "field": "transcriptions",
                "offset": 1337
            }
        }

    def make_act_hit(self, act, ts, score=1.0):
        return {
            "_score": score,
            "_type": Act.INDEX_TYPE,
            "_id": str(act.id.hex),
            "_index": "acts",
            "inner_hits": {
                "transcriptions": {
                    "hits": {
                        "total": len(ts),
                        "hits": list(map(self.make_nested_transcription_hit, ts)),
                        "max_score": 1337,
                    }
                }
            }
        }

    def test_search_transcriptions_post(self):
        expected = Transcription.objects.filter(text="PARIS")
        results = search_transcriptions_post(self.build_es_response(
            list(map(self.make_transcription_hit, expected))
        ))
        self.assertCountEqual(results, expected)

    def test_search_acts_post(self):
        act = Act.objects.get(number="1")
        ts = Transcription.objects.filter(text__in=["PARIS", "ROY"], zone__image__path='img1')
        surf = Element.objects.get(name="Surface A").zone
        results = search_acts_post(self.build_es_response(
            [self.make_act_hit(act, ts), ]
        ))

        self.assertCountEqual(results, [act])
        self.assertTrue(hasattr(results[0], 'transcriptions'))
        self.assertTrue(hasattr(results[0], 'surfaces'))
        self.assertCountEqual(results[0].transcriptions_results, ts)
        self.assertCountEqual(results[0].surfaces, [surf])

    def test_search_transcriptions_filter_post(self):
        # Filter to only get transcriptions from volume 1
        unfiltered = Transcription.objects.filter(text="PARIS")
        expected = Transcription.objects.filter(text="PARIS", zone__image__path__in=['img1', 'img2', 'img3'])
        results = search_transcriptions_filter_post(self.build_es_response(
            list(map(self.make_transcription_hit, unfiltered))
        ), Element.objects.get(name="Volume 1").id)
        self.assertCountEqual(results, expected)
