from django.db import models
from arkindex.project.models import IndexableModel
from arkindex.project.polygon import PolygonField
from enumfields import EnumField, Enum
import requests
import urllib.parse
import logging

logger = logging.getLogger(__name__)


class ImageServer(models.Model):
    """
    A IIIF image server
    """
    name = models.CharField(max_length=250)
    url = models.URLField()

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.name

    def find_image(self, path, offline=False, width=None, height=None):
        """
        Lookup an image on server
        This is the preferred way to construct an image
        """
        img = None
        # Try direct access to path
        try:
            img = self.images.get(path=path)
        except Image.DoesNotExist:
            pass

        if img is None:
            # Try the url encoded path
            try:
                img = self.images.get(
                    path=urllib.parse.quote_plus(path),
                )
            except Image.DoesNotExist:
                pass

        # Support offline queries
        if offline:
            if img is None:
                img = self.images.create(path=path, width=width, height=height)
            return img

        # Check the source
        if not path.endswith('/'):
            path += '/'
        info_url = urllib.parse.urljoin(self.build_url(path), 'info.json')
        resp = requests.get(info_url, allow_redirects=True)
        resp.raise_for_status()
        data = resp.json()

        if img is None:
            # Use Image id from IIIF server and create image
            image_id = data.get('@id')
            assert image_id is not None, \
                'Missing image id in server response'
            assert image_id.startswith(self.url), \
                'Image id does not start with server url ({} vs. {})'.format(
                    image_id, self.url)
            image_path = image_id[len(self.url) + 1:]
            img = self.images.create(path=image_path)

        assert 'width' in data, 'Missing image width in server response'
        assert 'height' in data, 'Missing image height in server response'
        if img.width != data['width'] or img.height != data['height']:
            # Missing width or height data in image
            img.width = data['width']
            img.height = data['height']

        img.status = ImageStatus.Checked
        img.save()

        return img

    def build_url(self, path):
        """
        Build a full url on this server
        """
        base_url = self.url
        if not base_url.endswith('/'):
            base_url += '/'
        return urllib.parse.urljoin(base_url, urllib.parse.quote(path.lstrip('/')))


class ImageStatus(Enum):
    """
    Image validation status
    """
    Checked = "checked"
    Unchecked = "unchecked"
    Error = "error"


class Image(IndexableModel):
    """
    A document image
    """
    server = models.ForeignKey(ImageServer, on_delete=models.CASCADE, related_name='images')
    path = models.URLField()
    width = models.PositiveIntegerField(default=0)
    height = models.PositiveIntegerField(default=0)
    status = EnumField(ImageStatus, default=ImageStatus.Unchecked, max_length=50)

    class Meta:
        unique_together = (
            ('server', 'path'),
        )

    @property
    def url(self):
        return self.server.build_url(self.path)

    def get_thumbnail_url(self, max_width=200, max_height=None):
        if max_width is None and max_height is None:
            resize = "full"
        else:
            resize = "{0},{1}".format(max_width or '', max_height or '')
        return "{0}/full/{1}/0/default.jpg".format(self.url.rstrip('/'), resize)

    def __str__(self):
        return '{} - {}'.format(self.id, self.url)


class Zone(IndexableModel):
    """
    A zone on an image
    """
    image = models.ForeignKey(Image, on_delete=models.CASCADE, related_name='zones')

    polygon = PolygonField()

    def __str__(self):
        return 'Zone {}'.format(self.polygon)

    @property
    def url(self):
        '''
        Build the IIIF crop url
        '''
        return urllib.parse.urljoin(
            self.image.url + '/',
            '{},{},{},{}/full/0/default.jpg'.format(
                self.polygon.x, self.polygon.y, self.polygon.width, self.polygon.height),
        )
