from django.db import models
from django.utils.functional import cached_property
import collections
import re

REGEX_SPLIT = re.compile(r'\((\d+),\s?(\d+)\)')


class Point(object):
    """
    Describe a point in the space.
    """
    def __init__(self, x, y):
        self.x = int(x)
        self.y = int(y)

    def __hash__(self):
        return hash((self.x, self.y))

    def __repr__(self):
        return 'Point({0.x},{0.y})'.format(self)

    def __str__(self):
        return '({0.x},{0.y})'.format(self)

    def __eq__(self, other):
        if isinstance(other, collections.abc.Sequence):
            return (self.x, self.y) == other
        return (isinstance(other, self.__class__) and
                self.x == other.x and
                self.y == other.y)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return (isinstance(other, self.__class__) and
                self.x <= other.x and
                self.y <= other.y)


class Polygon(collections.abc.MutableSequence):
    '''
    A hashable Polygon in-memory
    '''
    def __init__(self, points):
        assert len(points) > 0
        if not all(isinstance(point, Point) for point in points):
            points = [Point(*p) for p in points]

        # Clone & Close the polygon
        self.points = list(points)
        if self.points[0] != self.points[-1]:
            self.points.append(points[0])

    @staticmethod
    def from_dict(data):
        '''
        Helper to build an instance from dict data:
        * from a direct polygon as list of tuples
        * from x,y + width,height
        '''
        assert isinstance(data, dict)

        # Direct usage
        if 'polygon' in data and isinstance(data['polygon'], list):
            return Polygon(data['polygon'])

        # Build from coords
        for k in ('x', 'y', 'width', 'height'):
            assert k in data, 'Missing key {}'.format(k)
            assert data[k] is not None, 'Value of {} is None'.format(k)
            assert isinstance(data[k], int) or isinstance(data[k], float), \
                'Value of {} must be int or float'.format(k)

        return Polygon([
            (data['x'], data['y']),
            (data['x'] + data['width'], data['y']),
            (data['x'] + data['width'], data['y'] + data['height']),
            (data['x'], data['y'] + data['height']),
            (data['x'], data['y']),
        ])

    @staticmethod
    def from_str(data):
        assert isinstance(data, str)
        return Polygon([
            (float(x), float(y))
            for x, y in REGEX_SPLIT.findall(data)
        ])

    @staticmethod
    def from_coords(x, y, width, height):
        return Polygon.from_dict({'x': x, 'y': y, 'width': width, 'height': height})

    def __repr__(self):
        return "Polygon({})".format(', '.join(map(str, self.points)))

    def __eq__(self, other):
        other_points = other.points if hasattr(other, 'points') else other
        if len(self.points) != len(other_points):
            return False
        return all(
            x == y
            for x, y in zip(self.points, other_points)
        ) or all(
            x == y
            for x, y in zip(self.points, reversed(other_points))
        )

    def __hash__(self):
        return hash(tuple(self.points))

    def __getitem__(self, i):
        return self.points[i]

    def __setitem__(self, i, value):
        self.points[i] = value

    def __delitem__(self, i):
        del self.points[i]

    def __len__(self):
        return len(self.points)

    def insert(self, i, value):
        self.points.insert(i, value)

    def serialize(self):
        '''
        Used to save data in DB
        '''
        return [(p.x, p.y) for p in self.points]

    @cached_property
    def x(self):
        return min([point.x for point in self.points])

    @cached_property
    def y(self):
        return min([point.y for point in self.points])

    @cached_property
    def width(self):
        return max([point.x for point in self.points]) - self.x

    @cached_property
    def height(self):
        return max([point.y for point in self.points]) - self.y

    @property
    def center(self):
        return Point(
            self.x + self.width / 2.0,
            self.y + self.height / 2.0,
        )


class PolygonField(models.Field):
    """
    Field to store a polygon; needs at least three set of points
    """
    def db_type(self, connection):
        return 'polygon'

    def to_python(self, values):
        if values is None:
            return None

        if isinstance(values, str):
            polygon = Polygon.from_str(values)

        if not isinstance(values, collections.Iterable):
            raise TypeError("Value {} is not iterable".format(values))

        return polygon

    def from_db_value(self, value, expression, connection):
        if value is None:
            return

        return Polygon.from_str(value)

    def get_prep_value(self, values):
        if values is None:
            return
        if isinstance(values, list) or isinstance(values, tuple):
            if len(values) < 3:
                raise ValueError("Needs at minimum 3 points")
            polygon = Polygon(values)
        elif isinstance(values, str):
            polygon = Polygon.from_str(values)
        elif isinstance(values, Polygon):
            polygon = values
        else:
            raise Exception('Not supported polygon input')

        return ', '.join(map(str, polygon.points)) if values else None

    def get_prep_lookup(self, lookup_type, value):
        raise NotImplementedError(self)


@PolygonField.register_lookup
class PolygonContains(models.Lookup):
    lookup_name = 'contains'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "%s @> %s::polygon" % (lhs, rhs), params


@PolygonField.register_lookup
class PolygonIn(models.Lookup):
    lookup_name = 'in'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "%s <@ %s::polygon" % (lhs, rhs), params


@PolygonField.register_lookup
class PolygonExact(models.Lookup):
    lookup_name = 'exact'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "%s ~= %s::polygon" % (lhs, rhs), params
